'''module SendRx'''
from os import path
import time
# import sys
import json
# import math
import struct
from binascii import hexlify
from pyZynoUnifiedDrivers.miva.module_utils import crc32c


def get_dilute_volume(protocol):
    '''get dilute volume'''
    dilute_volume = None
    if protocol['content']['drug'] is not None:
        dilute_volume_str = protocol['content']['drug']['content']['diluteVolume']['value']
        dilute_volume = float(dilute_volume_str)
    return dilute_volume


def get_drug_amount(protocol):
    '''get drug amount
       This function converts the drug amount unit to micro gram'''

    drug_amount = None
    if protocol['content']['drug'] is not None:
        drug_amount_str = protocol['content']['drug']['content']['drugAmount']['value']
        drug_amount_unit_str = protocol['content']['drug']['content']['drugAmount']['unit']
        drug_amount = float(drug_amount_str)
        if drug_amount_unit_str == 'mg':
            drug_amount = drug_amount * 1000
    return drug_amount


def get_patient_weight(protocol):
    patient_weight = 0
    switches = protocol['content']['program']['switches']
    if switches['rate']:
        rate_str = protocol['content']['program']['rate']['value']
        rate = float(rate_str)
        rate_unit = protocol['content']['program']['rate']['unit']
        # drug amount (mcg)
        drug_amount = get_drug_amount(protocol)
        # dilute volume (mL)
        dilute_volume = get_dilute_volume(protocol)
        # weight (kg)
        if rate_unit in ('mg/kg/min', 'mcg/kg/min'):
            rate_upper_guard = 135
            rate_lower_guard = 0
            if rate_unit == 'mg/kg/min':
                weight_upper_limit = rate_upper_guard * drug_amount \
                            / (rate * 60 * dilute_volume * 1000)
                weight_lower_limit = rate_lower_guard * drug_amount \
                            / (rate * 60 * dilute_volume * 1000)
            if rate_unit == 'mcg/kg/min':
                weight_upper_limit = rate_upper_guard * drug_amount \
                            / (rate * 60 * dilute_volume)
                weight_lower_limit = rate_lower_guard * drug_amount \
                            / (rate * 60 * dilute_volume)
            
            patient_weight = input("    Input Patient Weight ({0:3.1f} - {1:3.1f} kg): "\
                            .format(weight_lower_limit, weight_upper_limit)) \
                            or (weight_upper_limit - weight_lower_limit) / 2
            try:
                patient_weight = float(patient_weight)
            except ValueError:
                patient_weight = 0

            if patient_weight > weight_upper_limit:
                patient_weight = weight_upper_limit
            elif patient_weight < weight_lower_limit:
                patient_weight = weight_lower_limit
            print('    patient weight = {:5.2f} kg'.format(patient_weight))
    return float(patient_weight)


def parse_rx_hex(rx_hex):
    '''parse rx hex'''
    rx_size = 512
    # convert rx_hex to rx_bytes
    rx_bytes = bytes.fromhex(rx_hex)
    print('rx size: {} bytes'.format(len(rx_bytes)))
    print('==')
    if len(rx_bytes) != rx_size:
        print('wrong rx size: {0} bytes (expected {1} bytes)'.format(len(rx_bytes), rx_size))
        return None
    #
    # 0:80 80-byte [Protocol]
    protocol_bytes = rx_bytes[0:80]
    protocol_hex = protocol_bytes.hex()
    print(protocol_hex)
    print('==[Protocol] parsing complete==')
    # 80:84 4-byte [Switches]
    switches_bytes = rx_bytes[80:84]
    switches_hex = switches_bytes.hex()
    print(switches_hex)
    print('==[Switches] parsing complete==')
    # 84:88 4-byte [Protocol CRC]
    protocol_crc_bytes = rx_bytes[84:88]
    protocol_crc_hex = protocol_crc_bytes.hex()
    print(protocol_crc_hex)
    print('==[Protocol CRC] parsing complete==')
    # 88:90 2-byte [Auth Role]
    auth_role_bytes = rx_bytes[88:90]
    auth_role_hex = auth_role_bytes.hex()
    print(auth_role_hex)
    print('==[Auth Role] parsing complete==')
    # 90:92 2-byte [Rate Factor]
    rate_factor_bytes = rx_bytes[88:90]
    rate_factor_hex = rate_factor_bytes.hex()
    print(rate_factor_hex)
    print('==[Rate Factor] parsing complete==')
    # 92:93 1-byte [Concentration Modifiable]
    concentration_modifiable_bytes = rx_bytes[92:93]
    concentration_modifiable_hex = concentration_modifiable_bytes.hex()
    print(concentration_modifiable_hex)
    print('==[Concentration Modifiable] parsing complete==')
    # 93:94 1-byte [Infusion Mode]
    infusion_mode_bytes = rx_bytes[93:94]
    infusion_mode_hex = infusion_mode_bytes.hex()
    print(infusion_mode_hex)
    print('==[Infusion Mode] parsing complete==')
    # 94:95 1-byte [ID]
    id_bytes = rx_bytes[94:95]
    id_hex = id_bytes.hex()
    print(id_hex)
    print('==[ID] parsing complete==')
    # 95:96 1-byte [Label Pool ID]
    label_pool_id_bytes = rx_bytes[95:96]
    label_pool_id_hex = label_pool_id_bytes.hex()
    print(label_pool_id_hex)
    print('==[Label Pool ID] parsing complete==')
    # 96:97 1-byte [Label ID]
    label_id_bytes = rx_bytes[96:97]
    label_id_hex = label_id_bytes.hex()
    print(label_id_hex)
    print('==[Label ID] parsing complete==')
    # 97:98 1-byte [Rate Unit]
    rate_unit_bytes = rx_bytes[97:98]
    rate_unit_hex = rate_unit_bytes.hex()
    print(rate_unit_hex)
    print('==[Rate Unit] parsing complete==')
    # 98:99 1-byte [Concentration Unit]
    concentration_unit_bytes = rx_bytes[98:99]
    concentration_unit_hex = concentration_unit_bytes.hex()
    print(concentration_unit_hex)
    print('==[Concentration Unit] parsing complete==')
    # 99:109 10-byte [Protocol Name]
    protocol_name_bytes = rx_bytes[99:109]
    protocol_name_hex = protocol_name_bytes.hex()
    print(protocol_name_hex)
    print('==[Protocol Name] parsing complete==')
    # 109:119 10-byte [Drug Name]
    drug_name_bytes = rx_bytes[109:119]
    drug_name_hex = drug_name_bytes.hex()
    print(drug_name_hex)
    print('==[Drug Name] parsing complete==')
    # 119:224 105-byte [Drug Components]
    drug_components_bytes = rx_bytes[119:224]
    drug_components_hex = drug_components_bytes.hex()
    print(drug_components_hex)
    print('==[Drug Components] parsing complete==')
    # 224:252 28-byte 0xFF
    #
    # 252:256 4-byte CRC
    crc_bytes = rx_bytes[252:256]
    crc_hex = crc_bytes.hex()
    print(crc_hex)
    print('==[CRC] parsing complete==')
    # 256:257 1-byte [Protocol Index]
    protocol_index_bytes = rx_bytes[256:257]
    protocol_index_hex = protocol_index_bytes.hex()
    print(protocol_index_hex)
    print('==[Protocol Index] parsing complete==')
    # 257:258 1-byte [View Index]
    view_index_bytes = rx_bytes[257:258]
    view_index_hex = view_index_bytes.hex()
    print(view_index_hex)
    print('==[View Index] parsing complete==')
    # 258:259 1-byte [Label Index]
    label_index_bytes = rx_bytes[258:259]
    label_index_hex = label_index_bytes.hex()
    print(label_index_hex)
    print('==[Label Index] parsing complete==')


    # 259:512 253-byte 0xFF
    #
def encrypt_rx(rx_path):
    rx = SendRx()
    rx.load(rx_path)
    rx_hex = rx.get_rx_hex()
    #
    parse_rx_hex(rx_hex)
    #
    rx_name = rx.send_rx['sendRxProtocol']['content']['name'].strip(' \t\r\n\0')
    # replace '\' or '/' or ' ' charactor to '' or '_' in the [rx_name]
    rx_name = rx_name.replace('\\', '')
    rx_name = rx_name.replace('/', '')
    rx_name = rx_name.replace(' ', '_')
    #
    rx_id = str(rx.send_rx['sendRxProtocol']['id'])
    time_stamp = str(int(time.time() * 1000))
    hex_file_path = 'SendRx-[' + rx_name + ']-[' + rx_id + ']@' \
            +time_stamp + '.hex'
    # write hex file
    file = open(hex_file_path, "w")
    file.write(rx_hex)
    file.close()
    print('Rx hex file is saved to: [{}]'.format(hex_file_path))
    # print('rx_hex = \n{}'.format(rx_hex))


class SendRx:
    
    PROTOCOL_UNION_SIZE = 80
    
    '''Send Rx class'''

    def __init__(self):
        self.send_rx = None
        self.pump_sn = ''
        self.library_crc = ''
        self.library_number = 0
        self.protocol_index = 0
        self.view_index = 255
        self.label_index = 255
        self.protocol = None
        self.weight = None

    def load(self, send_rx_path):
        '''load send rx'''
        # send_rx_json = ""
        # read json file
        file = open(send_rx_path, "r")
        send_rx_json = file.read()
        file.close()
        # convert JSON to Python Dict
        self.send_rx = json.loads(send_rx_json)
        # serialNumber: Pump Serial Number (should match the pump S/N)        
        self.pump_sn = self.send_rx['serialNumber']
        # libraryCrc: [crc] in library JSON
        self.library_crc = self.send_rx['libraryCrc']
        # libraryNumber: [id] in library JSON
        self.library_number = self.send_rx['libraryNumber']
        # protocolIndex: index of [extractedProtocols] list in library JSON
        self.protocol_index = self.send_rx['protocolIndex']
        # viewIndex: index of [views] list in library JSON (255: use [defaultViews] instead)
        self.view_index = self.send_rx['viewIndex']
        # labelIndex: index of [labelSets] list in library JSON (255: No Label)
        self.label_index = self.send_rx['labelIndex']
        # send rx protocol
        self.protocol = self.send_rx['sendRxProtocol']

    def get_pump_sn(self):
        '''get pump sn'''
        return self.pump_sn

    def get_library_crc(self):
        '''get library crc'''
        return self.library_crc

    def get_library_number(self):
        '''get library number'''
        return self.library_number

    def get_protocol_index(self):
        '''get protocol index'''
        return self.protocol_index

    def get_view_index(self):
        '''get view index'''
        return self.view_index

    def get_label_index(self):
        '''get label index'''
        return self.label_index

    def get_protocol(self):
        '''get protocol'''
        return self.protocol
        
    def get_weight(self):
        '''get weight'''
        return self.weight
        
    def get_rx_hex(self):
        '''get sendRx hex string'''
        # Get the bytes array
        # 0:80 80-byte [Protocol]
        protocol_bytes = self.get_protocol_bytes(self.protocol)
        # 80:84 4-byte [Switches]
        switches_bytes = self.get_switches_bytes(self.protocol)
        # 84:88 4-byte [Protocol CRC]
        protocol_crc_bytes = self.get_protocol_crc_bytes(self.protocol)
        # 88:90 2-byte [Auth Role]
        auth_role_bytes = int(0xFFFF).to_bytes(2, 'little')
        # 90:92 2-byte [Rate Factor]
        rate_factor_bytes = self.get_rate_factor_bytes(self.protocol)
        # 92:93 1-byte [Concentration Modifiable]
        concentration_modifiable_bytes = int(0).to_bytes(1, 'little')
        # 93:94 1-byte [Infusion Mode]
        infusion_mode_bytes = self.get_infusion_mode_bytes(self.protocol)
        # 94:95 1-byte [ID]
        id_bytes = int(0xFF).to_bytes(1, 'little')
        # 95:96 1-byte [Label Pool ID]
        label_pool_id_bytes = int(0xFF).to_bytes(1, 'little')
        # 96:97 1-byte [Label ID]
        label_id_bytes = int(0).to_bytes(1, 'little')
        # 97:98 1-byte [Rate Unit]
        rate_unit_bytes = self.get_rate_unit_bytes(self.protocol)
        # 98:99 1-byte [Concentration Unit]
        concentration_unit_bytes = self.get_concentration_unit_bytes(self.protocol)
        # 99:109 10-byte [Protocol Name]
        protocol_name_bytes = self.get_protocol_name_bytes(self.protocol)
        # 109:119 10-byte [Drug Name]
        drug_name_bytes = self.get_drug_name_bytes(self.protocol)
        # 119:224 105-byte [Drug Components]
        drug_components_bytes = self.get_drug_components_bytes(self.protocol)
        #
        send_rx_bytes = b''        
        send_rx_bytes += protocol_bytes
        send_rx_bytes += switches_bytes 
        send_rx_bytes += protocol_crc_bytes
        send_rx_bytes += auth_role_bytes
        send_rx_bytes += rate_factor_bytes
        send_rx_bytes += concentration_modifiable_bytes
        send_rx_bytes += infusion_mode_bytes
        send_rx_bytes += id_bytes
        send_rx_bytes += label_pool_id_bytes
        send_rx_bytes += label_id_bytes
        send_rx_bytes += rate_unit_bytes
        send_rx_bytes += concentration_unit_bytes
        send_rx_bytes += protocol_name_bytes
        send_rx_bytes += drug_name_bytes
        send_rx_bytes += drug_components_bytes
        # Append 0xFF
        while len(send_rx_bytes) < 256 - 4:
            send_rx_bytes += int(0xFF).to_bytes(1, 'little')
        send_rx_hex = send_rx_bytes.hex().upper()
        crc_bytes = crc32c(0, send_rx_hex).to_bytes(4, 'little')
        send_rx_bytes += crc_bytes
        #
        # 256:227 1-byte [Protocol Index]
        if int(self.protocol_index) >= 0:
            protocol_index_bytes = int(self.protocol_index).to_bytes(1, 'little')
        else:
            protocol_index_bytes = int(0xFF).to_bytes(1, 'little')
        # 257:258 1-byte [View Index]
        view_index_bytes = int(self.view_index).to_bytes(1, 'little')
        # 258:259 1-byte [Label Index]
        label_index_bytes = int(self.label_index).to_bytes(1, 'little')
        #
        send_rx_bytes += protocol_index_bytes
        send_rx_bytes += view_index_bytes
        send_rx_bytes += label_index_bytes
        # Append 0x00
        while len(send_rx_bytes) < 512:
            send_rx_bytes += int(0x00).to_bytes(1, 'little')
        # Convert byte array to hex string
        send_rx_hex = send_rx_bytes.hex().upper()
        # send_rx_hex = '000070420000a04100000040000000003c000000000070413c0000000000000000000000000000000000000000005243ffff7f7f00000000000000000000c84200007a4400000000ffffffffffffffff77004000ca46961cffff64000001ffff00000050726f746f636f6c200058595a000000000000006162630000000000000000000000000000000000302e3035206d672f6d6c00000000006566670000000000000000000000000000000000302e3035206d672f6d6c00000000000000000000000000000000000000000000000000000000000000000000000000000000ffffffffffffffffffffffffffffffffffffffffffffffffffffffffbbb5984305ffff00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000'
        return send_rx_hex
        
    def get_infusion_mode_bytes(self, protocol):
        '''Get Infusion Mode Bytes'''
        deliveryMode = protocol['content']['deliveryMode']
        delivery_mode_bytes = int(0)
        # Continuous Mode
        if deliveryMode == 'continuousInfusion':
            delivery_mode_bytes = int(0)
        # Bolus Mode
        if deliveryMode == 'bolusInfusion':
            delivery_mode_bytes = int(1)
        # Intermittent Mode
        if deliveryMode == 'intermittentInfusion':
            delivery_mode_bytes = int(2)
        return delivery_mode_bytes.to_bytes(1, 'little')

    def get_concentration_unit_bytes(self, protocol):
        concentration_unit_bytes = int(0)
        drug_unit = ''
        if protocol['content']['drug'] is not None: 
            drug_unit = protocol['content']['drug']['content']['drugAmount']['unit']
        if drug_unit == 'mg':
            concentration_unit_bytes = int(0)
        if drug_unit == 'mcg':
            concentration_unit_bytes = int(1)
        return concentration_unit_bytes.to_bytes(1, 'little')

    def get_rate_unit_bytes(self, protocol):
        '''Get Rate Unit Bytes'''
        delivery_mode = protocol['content']['deliveryMode']
        rate_unit_bytes = int(0)
        # Continuous Mode
        if delivery_mode == 'continuousInfusion':
            rate_unit = protocol['content']['program']['rate']['unit']
            rate_unit_bytes = int(0)
            if rate_unit == 'mL/hr':
                rate_unit_bytes = int(0)
            if rate_unit == 'mg/min':
                rate_unit_bytes = int(1)
            if rate_unit == 'mg/kg/min':
                rate_unit_bytes = int(2)
            if rate_unit == 'mcg/min':
                rate_unit_bytes = int(3)
            if rate_unit == 'mcg/kg/min':
                rate_unit_bytes = int(4)
        # Bolus Mode
        if delivery_mode == 'bolusInfusion':
            rate_unit_bytes = int(0)
        # Intermittent Mode
        if delivery_mode == 'intermittentInfusion':
            rate_unit_bytes = int(0)
        return rate_unit_bytes.to_bytes(1, 'little')

    def get_protocol_crc_bytes(self, protocol):
        protocol_crc_bytes = int(0).to_bytes(4, 'little')
        if len(protocol['crc']) == 8:
            # the crc is crc32c
            protocol_crc_bytes = int(protocol['crc'], 16).to_bytes(4, 'little')
        if protocol['crc'].find('MD5--') != -1:
            # the crc is MD5
            md5_ascii = protocol['crc'][5:]
            md5_hex = hexlify(md5_ascii.encode()).decode()
            protocol_crc_bytes = crc32c(0, md5_hex).to_bytes(4, 'little')
            # protocol_crc_bytes = int(1).to_bytes(4, 'little')
        return protocol_crc_bytes

    def get_drug_components_bytes(self, protocol):
        drug_components = None
        drug_components_bytes = b''
        if protocol['content']['drug'] is not None:
            drug_components = protocol['content']['drug']['content']['components']
            if drug_components != []:
                # print('drug_components = {}'.format(drug_components))
                for each in drug_components:
                    component_name = each['name']
                    component_name = component_name.encode()                
                    while len(component_name) < 20:
                        component_name += int(0x0).to_bytes(1, 'little')

                    component_concentration = each['concentration']
                    component_concentration = component_concentration.encode()
                    while len(component_concentration) < 15:
                        component_concentration += int(0x0).to_bytes(1, 'little')

                    drug_components_bytes += component_name
                    drug_components_bytes += component_concentration
            else:
                drug_components = 'No components'
                drug_components_bytes = 'No components'.encode()            
        while len(drug_components_bytes) < 105:
            drug_components_bytes += int(0).to_bytes(1, 'little')
        return drug_components_bytes

    def get_drug_name_bytes(self, protocol):
        '''Get Drug Name Bytes'''
        drug_name_bytes = int(0).to_bytes(1, 'little')
        if protocol['content']['drug'] is not None:
            drug_name_bytes = protocol['content']['drug']['content']['name']
            drug_name_bytes = drug_name_bytes.encode()
        while len(drug_name_bytes) < 10:
            drug_name_bytes += int(0).to_bytes(1, 'little')
        return drug_name_bytes

    def get_protocol_name_bytes(self, protocol):
        '''get protocol name bytes'''
        protocol_name = protocol['content']['name'].rstrip(' \t\r\n\0')
        if len(protocol_name) >= 9:
            protocol_name = protocol_name[0:9]
        protocol_name_bytes = protocol_name.encode()
        while len(protocol_name_bytes) < 10:
            protocol_name_bytes += int(0x0).to_bytes(1, 'little')
        return protocol_name_bytes

    def get_rate_factor_bytes(self, protocol):
        '''get flow rate calibration factor
           in continuous mode'''
        switches = protocol['content']['program']['switches']
        if switches['flowRateCalibrationFactor']:
            rate_factor_bytes = int(protocol['content']['program']['flowRateCalibrationFactor']['value'])        
        else:
            rate_factor_bytes = int(100)
        rate_factor_bytes = rate_factor_bytes.to_bytes(2, 'little')
        return rate_factor_bytes

    def get_switches_bytes(self, protocol):
        '''get switches of the protocol
           in continuous mode
           
           protocol - python dict converted from JSON string
           switches - 32-bit (4-byte) integer
           
           Description: 
           If the parameter is used, the corresponding bit is set, 
           otherwise, the corresponding bit is cleared
        '''
        switches = protocol['content']['program']['switches']
        protocol_switches = int(0)
        if protocol['content']['deliveryMode'] == 'continuousInfusion':
            # Rate
            if switches['rate']:
                protocol_switches |= 1 << 0
            # Vtbi
            if switches['vtbi']:
                protocol_switches |= 1 << 1
            # Time
            if switches['time']:
                protocol_switches |= 1 << 3
            # Delay Start
            if switches['delayStart']:
                protocol_switches |= 1 << 4
            # KVO Rate
            if switches.get('kvoRate') != None and switches['kvoRate']:
                protocol_switches |= 1 << 5
            # Delay KVO Rate
            if switches['delayKvoRate']:
                protocol_switches |= 1 << 6
        
            rate_unit = protocol['content']['program']['rate']['unit']
            if rate_unit is not None:
                # Concentration
                if rate_unit != 'mL/hr':
                    protocol_switches |= 1 << 7
                # Weight
                if rate_unit == 'mg/kg/min' or rate_unit == 'mcg/kg/min':
                    protocol_switches |= 1 << 10
            # Drug Amount
            if protocol['content']['drug'] is not None:
                protocol_switches |= 1 << 22
        if protocol['content']['deliveryMode'] == 'bolusInfusion':
            # Basal Rate
            if switches['basalRate']:
                protocol_switches |= 1 << 0
            # VTBI
            if switches['vtbi']:
                protocol_switches |= 1 << 1
            # Loading Dose Amount
            if switches['loadingDoseAmount']:
                protocol_switches |= 1 << 2
            # Delay Start
            if switches['delayStart']:
                protocol_switches |= 1 << 4
            # Auto Bolus Amount
            if switches['autoBolusAmount']:
                protocol_switches |= 1 << 5
            # Bolus Interval
            if switches['bolusInterval']:
                protocol_switches |= 1 << 6
            # Demand Bolus Amount
            if switches['demandBolusAmount']:
                protocol_switches |= 1 << 7
            # Lockout Time
            if switches['lockoutTime']:
                protocol_switches |= 1 << 8
            # Delay KVO Rate
            if switches['delayKvoRate']:
                protocol_switches |= 1 << 10
            # Max Amount Per Hour
            if switches['maxAmountPerHour']:
                protocol_switches |= 1 << 11
            # Max Amount Per Interval
            if switches['maxAmountPerInterval']:
                protocol_switches |= 1 << 12
            # Clinician Dose
            if switches['clinicianDose']:
                protocol_switches |= 1 << 13
            # Concentration
            basal_rate_unit = protocol['content']['program']['basalRate']['unit']
            if basal_rate_unit != 'mL/hr':
                protocol_switches |= 1 << 14
            if protocol['content']['drug'] is not None:
                # Drug Amount
                # protocol_switches |= 1 << 15
                # Dilute Volume
                # protocol_switches |= 1 << 16
                pass
            # Weight
            if basal_rate_unit == 'mg/kg/min' or basal_rate_unit == 'mcg/kg/min':
                protocol_switches |= 1 << 17
            # Drug Amount
            if protocol['content']['drug'] is not None:
                protocol_switches |= 1 << 22
        if protocol['content']['deliveryMode'] == 'intermittentInfusion':
            # Dose Rate
            if switches['doseRate']:
                protocol_switches |= 1 << 0
            # Dose VTBI
            if switches['doseVtbi']:
                protocol_switches |= 1 << 1
            # Total Time
            if switches['totalTime']:
                protocol_switches |= 1 << 3
            # Interval Time
            if switches['intervalTime']:
                protocol_switches |= 1 << 4
            # Delay Start
            if switches['delayStart']:
                protocol_switches |= 1 << 5
            # Intermittent KVO Rate
            if switches['intermittentKvoRate']:
                protocol_switches |= 1 << 6
            # KVO Rate
            if switches.get('kvoRate') != None and switches['kvoRate']:
                protocol_switches |= 1 << 7
            # Delay KVO Rate
            if switches['delayKvoRate']:
                protocol_switches |= 1 << 8
            # Max Amount Per Hour
            if switches['maxAmountPerHour']:
                protocol_switches |= 1 << 9
            # Max Amount Per Interval
            if switches['maxAmountPerInterval']:
                protocol_switches |= 1 << 10
            # Concentration
            dose_rate_unit = protocol['content']['program']['doseRate']['unit']
            if dose_rate_unit != 'mL/hr':
                protocol_switches |= 1 << 11
            if protocol['content']['drug'] is not None:
                # Drug Amount
                # protocol_switches |= 1 << 12
                # Dilute Volume
                # protocol_switches |= 1 << 13
                pass
            # Weight
            if dose_rate_unit == 'mg/kg/min' or dose_rate_unit == 'mcg/kg/min':
                protocol_switches |= 1 << 14
            # Drug Amount
            if protocol['content']['drug'] is not None:
                protocol_switches |= 1 << 22
        return protocol_switches.to_bytes(4, 'little')

    def get_protocol_bytes(self, protocol):
        '''get protocol bytes
           get a byte array of the protocol (80-byte)
           protocol - python dict converted from JSON string'''
        deliveryMode = protocol['content']['deliveryMode']
        switches = protocol['content']['program']['switches']
        # Continuous Mode
        if deliveryMode == 'continuousInfusion':
            # 0:4 Rate 4-byte
            rate_bytes = float(0)
            if switches['rate']:
                rate_bytes = float(protocol['content']['program']['rate']['value'])
            else:
                rate_bytes = float(0)
            rate_bytes = bytearray(struct.pack("f", rate_bytes))        
            # 4:8 VTBI 4-byte
            if switches['vtbi']:
                vtbi_bytes = float(protocol['content']['program']['vtbi']['value'])
            else:
                vtbi_bytes = float(0)
            vtbi_bytes = bytearray(struct.pack("f", vtbi_bytes))
            # 8:12 Loading Dose 4-byte
            loading_dose_bytes = float(0)
            loading_dose_bytes = bytearray(struct.pack("f", loading_dose_bytes))
            # 12:16 Time 4-byte
            if switches['time']:
                time_bytes = int(protocol['content']['program']['time']['value']) * 60
            else:
                time_bytes = int(0)
            time_bytes = time_bytes.to_bytes(4, 'little')
            # 16:20 Delay Start 4-byte
            if switches['delayStart']:
                delay_start_bytes = int(protocol['content']['program']['delayStart']['value'])
                # convert minutes to seconds
                delay_start_bytes = delay_start_bytes * 60
            else:
                delay_start_bytes = int(0)
            delay_start_bytes = delay_start_bytes.to_bytes(4, 'little')
            # 20:24 KVO Rate 4-byte
            if switches.get('kvoRate') != None and switches['kvoRate']:
                kvo_rate_bytes = float(protocol['content']['program']['kvoRate']['value'])
            else:
                kvo_rate_bytes = float(0)
            kvo_rate_bytes = bytearray(struct.pack("f", kvo_rate_bytes))        
            # 24:28 Delay KVO Rate 4-byte
            if switches['delayKvoRate']:
                delay_kvo_rate_bytes = float(protocol['content']['program']['delayKvoRate']['value'])
            else:
                delay_kvo_rate_bytes = float(0)
            delay_kvo_rate_bytes = bytearray(struct.pack("f", delay_kvo_rate_bytes))
            # 28:32 Concentration 4-byte
            if protocol['content']['drug'] is not None:
                drug_amount_bytes = float(protocol['content']['drug']['content']['drugAmount']['value'])
                solvent_volume_bytes = float(protocol['content']['drug']['content']['diluteVolume']['value'])
                concentration_bytes = drug_amount_bytes / solvent_volume_bytes
                concentration_bytes = bytearray(struct.pack("f", concentration_bytes))
            else:
                concentration_bytes = bytearray(struct.pack("f", 0))
            # 32:36 Drug Amount 4-byte
            if protocol['content']['drug'] is not None:
                drug_amount_bytes = float(protocol['content']['drug']['content']['drugAmount']['value'])
                drug_amount_bytes = bytearray(struct.pack("f", drug_amount_bytes))
            else:
                drug_amount_bytes = bytearray(struct.pack("f", 0))
            # 36:40 Solvent Volume 4-byte
            if protocol['content']['drug'] is not None:
                solvent_volume_bytes = float(protocol['content']['drug']['content']['diluteVolume']['value'])
                solvent_volume_bytes = bytearray(struct.pack("f", solvent_volume_bytes))
            else:
                solvent_volume_bytes = bytearray(struct.pack("f", 0))
            # 40:44 Weight 4-byte
            if self.weight is None: 
                self.weight = get_patient_weight(protocol)
            weight_bytes = bytearray(struct.pack("f", self.weight))

            protocol_bytes = b''
            protocol_bytes += rate_bytes
            protocol_bytes += vtbi_bytes
            protocol_bytes += loading_dose_bytes
            protocol_bytes += time_bytes
            protocol_bytes += delay_start_bytes
            protocol_bytes += kvo_rate_bytes
            protocol_bytes += delay_kvo_rate_bytes
            protocol_bytes += concentration_bytes
            protocol_bytes += drug_amount_bytes
            protocol_bytes += solvent_volume_bytes
            protocol_bytes += weight_bytes

        # Bolus Mode
        if deliveryMode == 'bolusInfusion':
            # 0:4 Basal Rate 4-byte Float
            basal_rate_bytes = float(0)
            if switches['basalRate']:
                basal_rate_bytes = float(protocol['content']['program']['basalRate']['value'])
            else:
                basal_rate_bytes = float(0)
            basal_rate_bytes = bytearray(struct.pack("f", basal_rate_bytes))        
            # 4:8 VTBI 4-byte Float
            if switches['vtbi']:
                vtbi_bytes = float(protocol['content']['program']['vtbi']['value'])
            else:
                vtbi_bytes = float(0)
            vtbi_bytes = bytearray(struct.pack("f", vtbi_bytes))
            # 8:12 Loading Dose 4-byte Float
            if switches['loadingDoseAmount']:
                loading_dose_bytes = float(protocol['content']['program']['loadingDoseAmount']['value'])
            else:
                loading_dose_bytes = float(0)
            loading_dose_bytes = bytearray(struct.pack("f", loading_dose_bytes))
            # 12:16 Time 4-byte Integer
            time_bytes = int(0)
            time_bytes = time_bytes.to_bytes(4, 'little')
            # 16:20 Delay Start 4-byte Integer
            if switches['delayStart']:
                delay_start_bytes = int(protocol['content']['program']['delayStart']['value'])
                # convert minutes to seconds
                delay_start_bytes = delay_start_bytes * 60
            else:
                delay_start_bytes = int(0)
            delay_start_bytes = delay_start_bytes.to_bytes(4, 'little')
            # 20:24 Auto Bolus Amount 4-byte Float
            if switches['autoBolusAmount']:
                auto_bolus_amount_bytes = float(protocol['content']['program']['autoBolusAmount']['value'])
            else:
                auto_bolus_amount_bytes = float(0)
            auto_bolus_amount_bytes = bytearray(struct.pack("f", auto_bolus_amount_bytes))
            # 24:28 Auto Bolus Interval 4-byte Integer (Minute)
            if switches['bolusInterval']:
                auto_bolus_interval_bytes = int(protocol['content']['program']['bolusInterval']['value'])
                # convert minutes to seconds
                auto_bolus_interval_bytes = auto_bolus_interval_bytes * 60
            else:
                auto_bolus_interval_bytes = int(0)
            auto_bolus_interval_bytes = auto_bolus_interval_bytes.to_bytes(4, 'little')
            # 28:32 Demand Bolus Volume 4-byte Float
            if switches['demandBolusAmount']:
                demand_bolus_amount_bytes = float(protocol['content']['program']['demandBolusAmount']['value'])
            else:
                demand_bolus_amount_bytes = float(0)
            demand_bolus_amount_bytes = bytearray(struct.pack("f", demand_bolus_amount_bytes))
            # 32:36 Demand Bolus Lockout Time 4-byte Integer
            if switches['lockoutTime']:
                lockout_time_bytes = int(protocol['content']['program']['lockoutTime']['value'])
                # convert minutes to seconds
                lockout_time_bytes = lockout_time_bytes * 60
            else:
                lockout_time_bytes = int(0)
            lockout_time_bytes = lockout_time_bytes.to_bytes(4, 'little')
            # 36:40 KVO Rate 4-byte Float
            kvo_rate_bytes = float(0)
            kvo_rate_bytes = bytearray(struct.pack("f", kvo_rate_bytes))        
            # 40:44 Delay KVO Rate 4-byte Float
            if switches['delayKvoRate']:
                delay_kvo_rate_bytes = float(protocol['content']['program']['delayKvoRate']['value'])
            else:
                delay_kvo_rate_bytes = float(0)
            delay_kvo_rate_bytes = bytearray(struct.pack("f", delay_kvo_rate_bytes))
            # 44:48 Max Amount Per Hour 4-byte Float
            if switches['maxAmountPerHour']:
                max_amount_per_hour = float(protocol['content']['program']\
                        ['maxAmountPerHour']['value'])
            else:
                max_amount_per_hour = float(0)
                max_amount_per_hour = float(210)
            max_amount_per_hour_bytes = bytearray(struct.pack("f", max_amount_per_hour))
            # 48:52 Max Amount Per Interval 4-byte Float
            if switches['maxAmountPerInterval']:
                max_amount_per_interval = float(protocol['content']['program']\
                        ['maxAmountPerInterval']['value'])
                max_amount_per_interval_bytes = bytearray(struct.pack("f", max_amount_per_interval))
            else:
                max_amount_per_interval = float(0)
                max_amount_per_interval_bytes = bytearray.fromhex('ffff7f7f')
            # 52:56 Clinician Dose 4-byte Float
            if switches['clinicianDose']:
                clinician_dose_bytes = float(protocol['content']['program']['clinicianDose']['value'])
            else:
                clinician_dose_bytes = float(0)
            clinician_dose_bytes = bytearray(struct.pack("f", clinician_dose_bytes))
            # 56:60 Concentration 4-byte Float
            if protocol['content']['drug'] is not None:
                # drug_amount_bytes = float(protocol['content']['drug']['content']['drugAmount']['value'])
                # solvent_volume_bytes = float(protocol['content']['drug']['content']['diluteVolume']['value'])
                # concentration = drug_amount_bytes / solvent_volume_bytes
                # concentration_bytes = bytearray(struct.pack("f", concentration))
                concentration_bytes = bytearray(struct.pack("f", float(0)))
            else:
                concentration_bytes = bytearray(struct.pack("f", 0))
            # 60:64 Drug Amount 4-byte Float
            if protocol['content']['drug'] is not None:
                drug_amount_bytes = float(protocol['content']['drug']['content']['drugAmount']['value'])
                drug_amount_bytes = bytearray(struct.pack("f", drug_amount_bytes))
            else:
                drug_amount_bytes = bytearray(struct.pack("f", 0))
            # 64:68 Solvent Volume 4-byte Float
            if protocol['content']['drug'] is not None:
                solvent_volume_bytes = float(protocol['content']['drug']['content']['diluteVolume']['value'])
                solvent_volume_bytes = bytearray(struct.pack("f", solvent_volume_bytes))
            else:
                solvent_volume_bytes = bytearray(struct.pack("f", 0))
            # 68:72 Weight 4-byte Float
            weight_bytes = float(0)
            weight_bytes = bytearray(struct.pack("f", weight_bytes))

            protocol_bytes = b''
            protocol_bytes += basal_rate_bytes
            protocol_bytes += vtbi_bytes
            protocol_bytes += loading_dose_bytes
            protocol_bytes += time_bytes
            protocol_bytes += delay_start_bytes
            protocol_bytes += auto_bolus_amount_bytes
            protocol_bytes += auto_bolus_interval_bytes
            protocol_bytes += demand_bolus_amount_bytes
            protocol_bytes += lockout_time_bytes
            protocol_bytes += kvo_rate_bytes
            protocol_bytes += delay_kvo_rate_bytes
            protocol_bytes += max_amount_per_hour_bytes
            protocol_bytes += max_amount_per_interval_bytes
            protocol_bytes += clinician_dose_bytes
            protocol_bytes += concentration_bytes
            protocol_bytes += drug_amount_bytes
            protocol_bytes += solvent_volume_bytes
            protocol_bytes += weight_bytes

        # Intermittent Mode
        if deliveryMode == 'intermittentInfusion':
            # 0:4 Dose Rate 4-Byte Float
            dose_rate_bytes = float(0)
            if switches['doseRate']:
                dose_rate_bytes = float(protocol['content']['program']['doseRate']['value'])
            else:
                dose_rate_bytes = float(0)
            dose_rate_bytes = bytearray(struct.pack("f", dose_rate_bytes))        
            # 4:8 Dose VTBI 4-Byte Float
            if switches['doseVtbi']:
                dose_vtbi_bytes = float(protocol['content']['program']['doseVtbi']['value'])
            else:
                dose_vtbi_bytes = float(0)
            dose_vtbi_bytes = bytearray(struct.pack("f", dose_vtbi_bytes))
            # 8:12 Loading Dose 4-Byte Float
            loading_dose_bytes = float(0)
            loading_dose_bytes = bytearray(struct.pack("f", loading_dose_bytes))
            # 12:16 Total Time 4-Byte Integer (Minute)
            if switches['totalTime']:
                total_time_bytes = int(protocol['content']['program']['totalTime']['value']) * 60
            else:
                total_time_bytes = int(0)
            total_time_bytes = total_time_bytes.to_bytes(4, 'little')
            # 16:20 Interval Time 4-Byte Integer (Minute)
            if switches['intervalTime']:
                interval_time_bytes = int(protocol['content']['program']['intervalTime']['value']) * 60
            else:
                interval_time_bytes = int(0)
            interval_time_bytes = interval_time_bytes.to_bytes(4, 'little')
            # 20:24 Delay Start 4-Byte Integer (Minute)
            if switches['delayStart']:
                delay_start_bytes = int(protocol['content']['program']['delayStart']['value'])
                # convert minutes to seconds
                delay_start_bytes = delay_start_bytes * 60
            else:
                delay_start_bytes = int(0)
            delay_start_bytes = delay_start_bytes.to_bytes(4, 'little')
            # 24:28 Intermittent KVO Rate 4-Byte Float
            if switches['intermittentKvoRate']:
                intermittent_kvo_rate_bytes = float(protocol['content']['program']['intermittentKvoRate']['value'])
            else:
                intermittent_kvo_rate_bytes = float(0)
            intermittent_kvo_rate_bytes = bytearray(struct.pack("f", intermittent_kvo_rate_bytes))        
            # 28:32 KVO Rate 4-Byte Float
            if switches.get('kvoRate') != None and switches['kvoRate']:
                kvo_rate_bytes = float(protocol['content']['program']['kvoRate']['value'])
            else:
                kvo_rate_bytes = float(0)
            kvo_rate_bytes = bytearray(struct.pack("f", kvo_rate_bytes))        
            # 32:36 Delay KVO Rate 4-Byte Float
            if switches['delayKvoRate']:
                delay_kvo_rate_bytes = float(protocol['content']['program']['delayKvoRate']['value'])
            else:
                delay_kvo_rate_bytes = float(0)
            delay_kvo_rate_bytes = bytearray(struct.pack("f", delay_kvo_rate_bytes))
            # 36:40 Max Amount Per Hour 4-Byte Float
            if switches['maxAmountPerHour']:
                max_amount_per_hour = float(protocol['content']['program']\
                        ['maxAmountPerHour']['value'])
            else:
                max_amount_per_hour = float(0)
                max_amount_per_hour = float(210)
            max_amount_per_hour_bytes = bytearray(struct.pack("f", max_amount_per_hour))
            # 40:44 Max Amount Per Interval 4-Byte Float
            if switches['maxAmountPerInterval']:
                max_amount_per_interval = float(protocol['content']['program']\
                        ['maxAmountPerInterval']['value'])
                max_amount_per_interval_bytes = bytearray(struct.pack("f", max_amount_per_interval))
            else:
                max_amount_per_interval_bytes = float(0)
                max_amount_per_interval_bytes = bytearray.fromhex('ffff7f7f')
            # 44:48 Concentration 4-Byte Float
            if protocol['content']['drug'] is not None:
                # drug_amount_bytes = float(protocol['content']['drug']['content']['drugAmount']['value'])
                # solvent_volume_bytes = float(protocol['content']['drug']['content']['diluteVolume']['value'])
                # concentration = drug_amount_bytes/solvent_volume_bytes
                # concentration_bytes = bytearray(struct.pack("f", concentration))
                concentration_bytes = bytearray(struct.pack("f", float(0)))
            else:
                concentration_bytes = bytearray(struct.pack("f", 0))
            # 48:52 Drug Amount  4-Byte Float
            if protocol['content']['drug'] is not None:
                drug_amount_bytes = float(protocol['content']['drug']['content']['drugAmount']['value'])
                drug_amount_bytes = bytearray(struct.pack("f", drug_amount_bytes))
            else:
                drug_amount_bytes = bytearray(struct.pack("f", 0))
            # 52:56 Solvent Volume 4-Byte Float
            if protocol['content']['drug'] is not None:
                solvent_volume_bytes = float(protocol['content']['drug']['content']['diluteVolume']['value'])
                solvent_volume_bytes = bytearray(struct.pack("f", solvent_volume_bytes))
            else:
                solvent_volume_bytes = bytearray(struct.pack("f", 0))
            # 56:60 Weight 4-Byte Float
            weight_bytes = float(0)
            weight_bytes = bytearray(struct.pack("f", weight_bytes))

            protocol_bytes = b''
            protocol_bytes += dose_rate_bytes
            protocol_bytes += dose_vtbi_bytes
            protocol_bytes += loading_dose_bytes
            protocol_bytes += total_time_bytes
            protocol_bytes += interval_time_bytes
            protocol_bytes += delay_start_bytes
            protocol_bytes += intermittent_kvo_rate_bytes
            protocol_bytes += kvo_rate_bytes
            protocol_bytes += delay_kvo_rate_bytes
            protocol_bytes += max_amount_per_hour_bytes
            protocol_bytes += max_amount_per_interval_bytes
            protocol_bytes += concentration_bytes
            protocol_bytes += drug_amount_bytes
            protocol_bytes += solvent_volume_bytes
            protocol_bytes += weight_bytes
        # Padding with 0xFF
        while (len(protocol_bytes) < self.PROTOCOL_UNION_SIZE):
            protocol_bytes += int(0xFF).to_bytes(1, 'little')
        return protocol_bytes


def main():
    '''main function'''
    cmd = ''
    while cmd not in ['exit', 'quit']:
        cmd = input('>')
        if cmd.upper().rstrip(' \t\r\n\0') in ['HELP', '?']:
            print(' ER - Encrypt Rx JSON To Byte Array')
            print(' PR - Parse Rx Hex')
        # Encrypt Rx JSON To Byte Array
        if cmd.lower().rstrip(' \t\r\n\0') in ['er', 'encrypt rx']:
            # result = False
            print(' ER - Encrypt Rx JSON To Byte Array')
            rx_path = input('    Enter Rx path: ')
            if path.exists(rx_path):
                encrypt_rx(rx_path)
            else:
                if rx_path == '':
                    pass
                else:
                    print('Abort: path NOT exist')
        # Parse Rx Hex
        if cmd.lower().rstrip(' \t\r\n\0') in ['pr', 'parse rx']:
            # result = False
            print(' PR - Parse Rx Hex')
            rx_hex_path = input('    Enter Rx Hex file path: ')
            if path.exists(rx_hex_path):
                rx_hex = ""
                # read hex file
                file = open(rx_hex_path, "r")
                rx_hex = file.read()
                file.close()
                # parse librar hex
                parse_rx_hex(rx_hex)
            else:
                if rx_hex_path == '':
                    pass
                else:
                    print('Abort: path NOT exist')


if __name__ == "__main__":
    main()
    



'''Socket Server SSL Module'''
import sys
import time
import traceback
import socket
from socketserver import TCPServer, BaseRequestHandler
import ssl
from pyZynoUnifiedDrivers.miva.module_utils import download_mfg_32_bit, download_miva_8_bit, download_miva_32_bit
PRIVATE_KEY_PATH = './miva/server_key_store/private_key.pem'
SERVER_CERTIFICATE_PATH = './miva/server_key_store/certificate.crt'

_TIMEOUT = 0.025

SOCKET_BUFFER_SIZE = 4096


class SocketRequestHandlerSSL(BaseRequestHandler):
    """
    The request handler class for our server.
 
    It is instantiated once per connection to the server, and must
    override the handle() method to implement communication to the
    client.
    """
 
    def handle(self):
        try:
            if (self.server.is_on()):
                # self.request is the TCP socket connected to the client
                self.data = self.request.recv(SOCKET_BUFFER_SIZE)
                # print('self.data = {}'.format(self.data))
                if self.data == b'':
                    # break if client closes the socket
                    pass
                    # break            
                elif self.data == b'\r\n':
                    # ignore [\r\n]
                    pass
                    # continue
                else:
                    #
                    print('{}'.format(time.strftime('%Y-%m-%d %H:%M:%S', time.localtime())))
                    # print(self.data)
                    # just send back the same data, but upper-cased
                    # self.request.sendall(self.data + b'\r\n')
                    # self.request.sendall(b'server status: ' + str(self.server.is_on()).encode() + b'\r\n')
                    message = self.data.decode(errors='ignore').strip(' \t\r\n\0')
                    print('get: \'{0}\' from {1}'.format(message, self.client_address[0]))
                    # print('length = {}'.format(len(message)))
                    # print()
                    if self.server.miva != None:
                        if message.lower() in [':download:mfg32', ':flash:mfg32', ':install:mfg32']:
                            return_message = download_mfg_32_bit()
                            if return_message:
                                return_message = 'Success: Download [mfg 32-bit firmware] with [pyProgrammerCLI.exe]'
                            else:
                                return_message = 'Fail: Download [mfg 32-bit firmware] with [pyProgrammerCLI.exe]'
                        elif message.lower() in [':download:miva8', ':flash:miva8', ':install:miva8']:
                            return_message = download_miva_8_bit()
                            if return_message:
                                return_message = 'Success: Download [miva 8-bit firmware] with [FlashUtilCL.exe]'
                            else:
                                return_message = 'Fail: Download [miva 8-bit firmware] with [FlashUtilCL.exe]'
                        elif message.lower() in [':download:miva32', ':flash:miva32', ':install:miva32']: 
                            return_message = download_miva_32_bit()
                            if return_message:
                                return_message = 'Success: Download [miva 32-bit firmware] with [pyProgrammerCLI.exe]'
                            else:
                                return_message = 'Fail: Download [miva 32-bit firmware] with [pyProgrammerCLI.exe]'
                        else:
                            return_message = self.server.miva.query(message)
                        # print('return_message = {}'.format('\'\'' if return_message == '' else return_message))
                    else:
                        return_message = self.server.process_incomming_message(message)                        
                    if return_message != '':
                        #
                        data = return_message.strip(' \t\r\n\0').encode()
                        self.request.sendall(data)
                        print('sent: \'{0}\''.format(return_message))
                        print()
        except socket.timeout:
            traceback.print_exc()
        except UnicodeDecodeError:
            print()
            # print("{0}: {1}\n".format(sys.exc_info()[0], sys.exc_info()[1]))
            traceback.print_exc()
        except (ConnectionResetError, OSError):
            print()
            print('{0}: {1}\n'.format(sys.exc_info()[0], sys.exc_info()[1]))
        except:
            traceback.print_exc()
        finally: 
            self.request.close()
            return ''


class SocketServerSSL(TCPServer): 

    def __init__(self,
                 server_address,
                 RequestHandlerClass,
                 certfile,
                 keyfile,
                 ssl_version=ssl.PROTOCOL_TLSv1,
                 bind_and_activate=True,
                 miva=None):
        '''__init__'''
        TCPServer.__init__(self, server_address, RequestHandlerClass, bind_and_activate)
        self.certfile = certfile
        self.keyfile = keyfile
        self.ssl_version = ssl_version
        
        self.HOST, self.PORT = server_address
        self._is_on = False
        self.miva = miva     

    def get_request(self):
        newsocket, fromaddr = self.socket.accept()
        connstream = ssl.wrap_socket(newsocket,
                                 server_side=True,
                                 certfile=self.certfile,
                                 keyfile=self.keyfile,
                                 ssl_version=self.ssl_version)
        return connstream, fromaddr

    def start(self):
        '''start ssl server'''
        try:
            # print("Socket server started at [{}]"\
            # .format(time.strftime('%H:%M:%S', time.localtime())))
            self.set_on()
            self.serve_forever()
            print("Socket server stopped at [{}]"\
                    .format(time.strftime('%H:%M:%S', time.localtime())))
        except KeyboardInterrupt:
            print('Exiting: [Ctrl + C] is pressed')
            pass        
        except OSError:
            print('Exiting: [OSError] is captured')
            pass
        except:
            traceback.print_exc()
            pass
        finally:
            self.stop()
        
    def stop(self):
        '''stop'''
        self.shutdown()
        self._is_on = False
        # print("Socket server stopped at [{}]"\
        # .format(time.strftime('%H:%M:%S', time.localtime())))
    
    def set_on(self):
        '''stop'''
        self._is_on = True
        print('Socket server [{}] is listening on port [{}]'.format(self.HOST, self.PORT))    
        print()
        
    def is_on(self):
        '''is on'''
        return self._is_on
    
    def status(self):
        if self._is_on:
            print("Gateway Server: [On]")
            print("\t{0:5}: {1}".format('IP', self.HOST))
            print("\t{0:5}: {1}".format('Port', str(self.PORT)))
        else:
            print("Gateway Server: [Off]")

    def process_incomming_message(self, cmd):
        if cmd.lower().strip(' \t\r\n\0') == '*idn?':
            return_msg = 'Socket Server SSL ({})'.format(self.HOST)        
        else:
            return_msg = ''
        return return_msg

        
def main():
    '''main function'''
    try:
        hostname = socket.gethostname()
        local_ip = socket.gethostbyname(hostname)
        # local_ip = '192.168.0.19'
        # print("IP Address: {0}".format(local_ip))
    
        # HOST = '127.0.0.1'  # Standard loopback interface address (localhost)
        HOST = local_ip
        PORT = 443  # Port to listen on (non-privileged ports are > 1023)
        # Create the server, binding to localhost on port 65432
        with SocketServerSSL((HOST, PORT),
                             SocketRequestHandlerSSL,
                             certfile=SERVER_CERTIFICATE_PATH,
                             keyfile=PRIVATE_KEY_PATH,) as socket_server_ssl:
            socket_server_ssl.start()
    except KeyboardInterrupt:
        print('Exiting: [Ctrl + C] is pressed')
        pass


if __name__ == "__main__":
    main()



'''Socket Module'''
import re
import json
import sys
import time
import datetime
# from threading import Thread
from os import path  # , system
import socket
import copy
from socketserver import BaseRequestHandler, TCPServer
from pyZynoUnifiedDrivers.miva.module_library import get_parameter_list, get_parameter
from pyZynoUnifiedDrivers.miva.module_infusion import InfusionStatus
from pyZynoUnifiedDrivers.miva.module_utils import json_to_file
from pyZynoUnifiedDrivers.miva.module_utils import _UP_KEY, _DOWN_KEY, _INFO_KEY, _OK_KEY, _POWER_KEY
from pyZynoUnifiedDrivers.miva.module_utils import _RUN_KEY, _BOLUS_KEY, _SHORT_PRESS, _LONG_PRESS, _MID_PRESS

_TIMEOUT = 0.025

SOCKET_BUFFER_SIZE = 4096


class SocketRequestHandler(BaseRequestHandler):
    """
    The request handler class for our server.
 
    It is instantiated once per connection to the server, and must
    override the handle() method to implement communication to the
    client.
    """
 
    def handle(self):
        try:
            while (self.server.is_on()):
                # self.request is the TCP socket connected to the client
                self.data = self.request.recv(SOCKET_BUFFER_SIZE)
                if self.data == b'':
                    # break if client closes the socket
                    break            
                elif self.data == b'\r\n':
                    # ignore [\r\n]
                    continue
                else:
                    print('{} wrote: '.format(self.client_address[0]))
                    print(self.data)
                    # just send back the same data, but upper-cased
                    # self.request.sendall(self.data + b'\r\n')
                    # self.request.sendall(b'server status: ' + str(self.server.is_on()).encode() + b'\r\n')
                    if self.server.is_on():
                        cmd = self.data.decode()
                        cmd = cmd.strip(' \t\r\n\0')
                        msg = process_command(self.server, cmd)
                    else:
                        msg = 'server status: Off'
                    self.request.sendall(msg.encode() + b'\r\n')
        except (ConnectionResetError, OSError):
            print()
            print('{0}: {1}\n'.format(sys.exc_info()[0], sys.exc_info()[1]))
        finally: 
            self.request.close()


class SocketServer(TCPServer): 

    def __init__(self, server_address, RequestHandlerClass, miva=None, library=None, infusion=None, infusion_monitor=None, event_log_monitor=None):
        '''__init__'''
        TCPServer.__init__(self, server_address, RequestHandlerClass)
        self._is_on = False
        self.tcp_server = None
        self.HOST, self.PORT = server_address
        self.miva = miva
        self.library = library
        self.infusion = infusion
        self.infusion_monitor = infusion_monitor
        self.event_log_monitor = event_log_monitor

    def start(self):
        '''start'''
        try:
            # print("Socket server started at [{}]"\
            # .format(time.strftime('%H:%M:%S', time.localtime())))
            self.set_on()
            self.serve_forever()
            print("Socket server stopped at [{}]"\
                    .format(time.strftime('%H:%M:%S', time.localtime())))
        except KeyboardInterrupt:
            print('Exiting: [Ctrl + C] is pressed')
            pass        
        except OSError:
            print('Exiting: [OSError] is captured')
            pass
        finally:
            self.stop()
        
    def stop(self):
        '''stop'''
        self.shutdown()
        self._is_on = False
        # print("Socket server stopped at [{}]"\
        # .format(time.strftime('%H:%M:%S', time.localtime())))
    
    def set_on(self):
        '''stop'''
        self._is_on = True        

    def is_on(self):
        '''is on'''
        return self._is_on
    
    def status(self):
        if self._is_on:
            print("Socket Server: [On]")
            print("\t{0:5}: {1}".format('IP', self.HOST))
            print("\t{0:5}: {1}".format('Port', str(self.PORT)))
        else:
            print("Socket Server: [Off]")


def process_command(socket_server, cmd):
    return_msg = ''
    miva = socket_server.miva
    library = socket_server.library
    infusion = socket_server.infusion
    protocol = infusion.protocol
    infusion_monitor = socket_server.infusion_monitor
    event_log_monitor = socket_server.event_log_monitor
    if miva is None:
        return return_msg
    #
    # Pause [Event Log Monitor]
    if event_log_monitor.is_on():
        if not event_log_monitor.is_paused():
            event_log_monitor.pause()
    #
    # Disable [Event Log Monitor] Output
    infusion_monitor.disable_output()
    #
    cmd_original = cmd.strip(' \t\r\n\0')
    cmd = cmd.lower().strip(' \t\r\n\0')
    if re.match(r'\*idn\?', cmd):
        # Identifies Instrument Type and Software Version.
        # *IDN?
        make = 'Zyno Medical'
        platform = miva.read_platform()
        model = 'Nimbus' + ' ' + platform
        serial_number = miva.get_pump_sn()
        software_version = miva.get_firmware_version()
        if serial_number != '':
            return_msg = make + ',' + model + ',' + serial_number + ',' + software_version
            print(return_msg)
        else:
            print('Error: get serial number failed')
    # ==============================================
    elif re.match(r'(:)(even(t)?)(:)(\d+)(\?)', cmd):
        # Read Last [N] Event Log
        # :EVENt:<N>?
        # [r'] means this string is a regular expression
        # (:even(t)?)   : must contain [event:] or [even:] ([t] is optional)
        # (:)          : must contain [:]
        # (\d+)        : must contain 1 or more digit
        # (\?)         : must contain a question mark [?]
        re_match_result = re.match(r'(:)(even(t)?)(:)(\d+)(\?)', cmd)
        # the number of event log to print
        num_to_print = int(re_match_result[5])
        #
        event_logs = miva.read_multiple_event_log(num_to_print)
        event_logs_json = json.dumps(event_logs)
        print(json.dumps(event_logs, indent=4))
        print('     Event logs read: [{}]'.format(len(event_logs)))
        #
        return_msg = event_logs_json
    elif re.match(r'(:)(even(t)?)(:)(ind(ex)?)((:)(head|tail))?(\?)', cmd):
        # Read Event Log Indices
        # :EVENt:INDex[:{HEAD | TAIL }]?
        # (head|tail) : matches either [head] or [tail]
        event_log_index = miva.read_event_log_indices()
        event_log_index_tail = event_log_index['tail']
        event_log_index_head = event_log_index['head']
        if event_log_index_tail != '' and event_log_index_head != '':
            print('Event log tail = {}'.format(event_log_index_tail))
            print('Event log head = {}'.format(event_log_index_head))
        else:
            print('Error: get event log index failed')
        #
        re_match_result = re.match(r'(:)(even(t)?)(:)(ind(ex)?)((:)(head|tail))?(\?)', cmd)
        if re_match_result[9] is not None:
            return_msg = event_log_index[re_match_result[9]]
        else:
            return_msg = json.dumps(event_log_index)
    elif re.match(r'(:)(ser(ial)?)(\?)', cmd):
        # Read Pump Serial Number
        # :SERial?
        serial_number = miva.get_pump_sn()
        if serial_number != '':
            print('S/N: {}'.format(serial_number))
            return_msg = serial_number
        else:
            print('Error: get serial number failed')
    elif re.match(r'(:)(ser(ial)?)(\s)([a-z0-9]{8,8})', cmd):
        # Write Pump Serial Number
        # :SERial <serial_number>
        print('Enter NEW S/N: ', end='')
        re_match_result = re.match(r'(:)(ser(ial)?)(\s)([a-z0-9]{8,8})', cmd)
        serial_number = re_match_result[5].lower()
        if len(serial_number) == 8 and re.findall(r"[a-z0-9]{8,8}", serial_number) != []:
            status = miva.set_pump_sn(serial_number.upper())
            if status:
                # print('New S/N: {}'.format(serial_number))
                print("Write pump serial number  -  Done!")
            else:
                print('Error: set serial number failed')
        else:
            print('Error: invalid serial number \'{}\' '.format(serial_number), end='')
            print('Enter \'SN\' followed by 6-digit pump serial number ', end='')
            print('(ex: \'SN800009\')')
    elif re.match(r'(:)(mot(or)?)(:)(cal(ibrate)?)(:)(fact(or)?)(\?)', cmd):
        # Read [Motor Calibration Factor]
        # :CALibrate:FACTor?
        motor_calibration_factor = miva.get_motor_calibration_factor()
        if motor_calibration_factor is not None:
            print('Motor calibration factor: {}'.format(motor_calibration_factor))
            return_msg = str(motor_calibration_factor)
        else:
            print('Error: get [motor calibration factor] failed')
    elif re.match(r'(:)(sens(or)?)((:)(up|down))?((:)(ref(erence)?))(\?)', cmd):
        # Read Occlusion Sensor Reference Value
        # :SENSor[:{UP | DOWN}]:REFerence?
        re_match_result = re.match(r'(:)(sens(or)?)((:)(up|down))?((:)(ref(erence)?))(\?)', cmd)
        occlusion_thresholds = miva.get_occlusion_thresholds()
        if occlusion_thresholds != {}:
            print('Up stream threshold: {}'.format(occlusion_thresholds['up']))
            print('Dn stream threshold: {}'.format(occlusion_thresholds['down']))
        else:
            print('Error: get [occlusion thresholds] failed')
        #
        if re_match_result[6] is not None:
            return_msg = str(occlusion_thresholds[re_match_result[6]])
        else:
            return_msg = json.dumps(occlusion_thresholds)
    elif re.match(r'(:)(sens(or)?)((:)(up|down))?((:)(ref(erence)?))(\s)(\d+)((\s)(\d+))?', cmd):
        # Write Occlusion Sensor Reference Value
        # :SENSor[:{UP | DOWN}]:REFerence <up_number> [<down_number>]
        re_match_result = re.match(r'(:)(sens(or)?)((:)(up|down))?((:)(ref(erence)?))(\s)(\d+)((\s)(\d+))?', cmd)
        occlusion_thresholds = miva.get_occlusion_thresholds()
        if re_match_result[6] is not None and re_match_result[15] is None:
            # matched {UP | DOWN} and didn't match the second number
            # Write only one sensor reference
            occlusion_thresholds[re_match_result[6]] = int(re_match_result[12]) if 0 <= int(re_match_result[12]) <= 65535 else 65535
        elif re_match_result[6] is None and re_match_result[15] is not None:
            # Didn't match {UP | DOWN} and matched the second number
            # Write both sensor references
            occlusion_thresholds['up'] = int(re_match_result[12]) if 0 <= int(re_match_result[12]) <= 65535 else 65535
            occlusion_thresholds['down'] = int(re_match_result[15]) if 0 <= int(re_match_result[15]) <= 65535 else 65535
        # Write the sensor reference values back
        occlusion_thresholds = miva.set_occlusion_thresholds(occlusion_thresholds)
        if occlusion_thresholds != {}:
            print('Up threshold is set to: {}'.format(occlusion_thresholds['up']))
            print('Dn threshold is set to: {}'.format(occlusion_thresholds['down']))
        else:
            print('Error: set [occlusion thresholds] failed')
    elif re.match(r'(:)(vinf)(:)(tot(al)?)(\?)', cmd):
        # Read [Total Volume Infused]
        # :VINF:TOTal?
        # total_volume_infused = miva.get_total_volume_infused()
        product_life_volume = miva.get_product_life_volume()
        if product_life_volume is not None:
            # print('Total volume infused: {} mL'.format(total_volume_infused))
            print('Total volume infused: {:.2f} mL'.format(product_life_volume['volume']))
            return_msg = '{:.2f}'.format(product_life_volume['volume'])
        else:
            print('Error: get [Total Volume Infused] failed')
    elif re.match(r'(:)(batt(ery)?)(:)(volt(age)?)(\?)', cmd):
        # Read [Battery Voltage]
        # :BATTery:VOLTage?
        battery_voltage = miva.get_battery_voltage()
        if battery_voltage is not None:
            print('Battery voltage: {:.2f} V'.format(battery_voltage))
            return_msg = '{:.2f}'.format(battery_voltage)            
        else:
            print('Error: get [Battery Voltage] failed')
    elif re.match(r'(:)(mot(or)?)(:)(rot(ation)?)(:)(tot(al)?)(\?)', cmd):
        # Read [Motor Total Rotation]
        # :MOTor:ROTation:TOTal?
        product_life_volume = miva.get_product_life_volume()
        if product_life_volume != {}:
            print('Motor total rotation: {}'.format(product_life_volume['rotation']))
            return_msg = str(product_life_volume['rotation'])
        else:
            print('Error: get [Calibration Factor and Total Rotation] failed')
    elif re.match(r'(:)(pump)(:)(on)(:)(time)(\?)', cmd):
        # Read [Total Pump ON Time] (in seconds)
        # :PUMP:ON:TIME?
        total_pump_on_time = miva.get_total_pump_on_time()
        if total_pump_on_time != '':
            day = str(int(total_pump_on_time / 3600 / 24)).zfill(2)
            hour = str(int(total_pump_on_time / 3600 % 24)).zfill(2)
            minute = str(int(total_pump_on_time % 3600 / 60)).zfill(2)
            second = str(int(total_pump_on_time % 3600 % 60)).zfill(2)
            print('{0} day {1} hr {2} min {3} sec'.format(day, hour, minute, second))
            print("Read pump total ON time - Done!")
            return_msg = str(total_pump_on_time)
        else:
            print('Error: get [Battery Voltage] failed')
    elif re.match(r'(:)(batt(ery)?)(:)(cal(ibrate)?)(:)(fact(or)?)(\?)', cmd):
        # Read Battery Calibration Factor
        # :BATTery:CALibrate:FACTor?
        battery_calibration_and_voltage = miva.get_battery_calibration_and_voltage()
        if battery_calibration_and_voltage != {}:
            print('Battery calibration factor = {}'\
                    .format(battery_calibration_and_voltage['calibration']))
            print('Battery voltage = {:.2f} V'\
                    .format(battery_calibration_and_voltage['voltage']))
            return_msg = str(battery_calibration_and_voltage['calibration'])
        else:
            print('Error: get [Battery Voltage] failed')
    elif re.match(r'(:)(base)(:)(year)(\?)', cmd):
        # Read [Base Year] (2000)
        # :BASE:YEAR?
        base_year = miva.get_base_year()
        if base_year != '':
            print('Base year = {}'.format(base_year))
            return_msg = str(base_year)
        else:
            print('Error: get [Base Year] failed')
    elif re.match(r'(:)(sens(or)?)((:)(up|down))?(\?)', cmd):
        # Read Occlusion Sensor Value
        # :SENSor[:{UP | DOWN}]?
        re_match_result = re.match(r'(:)(sens(or)?)((:)(up|down))?(\?)', cmd)
        occlusion_sensor = miva.get_occlusion_sensor()
        if occlusion_sensor != {}:
            print('Up stream: {}'.format(occlusion_sensor['up']))
            print('Dn stream: {}'.format(occlusion_sensor['down']))
        else:
            print('Error: get [occlusion thresholds] failed')
        #
        if re_match_result[6] is not None:
            return_msg = str(occlusion_sensor[re_match_result[6]])
        else:
            return_msg = json.dumps(occlusion_sensor)
    elif re.match(r'(:)(lib(rary)?)((:)(name|ver(sion)?))?(\?)', cmd):
        # Read [Library Name and Version]
        re_match_result = re.match(r'(:)(lib(rary)?)((:)(name|ver(sion)?))?(\?)', cmd)
        library_name_version = miva.get_library_name_version()
        if library_name_version != '':
            print('Library name: [{}]'.format(library_name_version['name']))
            print('Library version: [{}]'.format(library_name_version['version']))
        else:
            print('Error: get [Library Name and Version] failed')
        #
        if re_match_result[6] is not None:
            if re_match_result[6] == 'name':
                return_msg = str(library_name_version['name'])
            else:
                return_msg = str(library_name_version['version'])
        else:
            return_msg = json.dumps(library_name_version)
    elif re.match(r'(:)(firm(ware)?)(:)(ver(sion)?)(\?)', cmd):
        # Read [Firmware Version]
        software_version = miva.get_firmware_version()
        if software_version != '':
            print('Firmware version: [{}]'.format(software_version))
            return_msg = software_version
        else:
            print('Error: get [Software Version] failed')
    # ==============================================
    elif re.match(r'(:)(key)(:)(info)((:)(shor(t)?|long))?', cmd):
        # Send [INFO] key command [Short | Long] Press
        re_match_result = re.match(r'(:)(key)(:)(info)((:)(shor(t)?|long))?', cmd)
        if re_match_result[7] is not None:
            if re_match_result[7] == 'long':
                return_text = miva.press_key(_INFO_KEY, _LONG_PRESS)
            else:
                return_text = miva.press_key(_INFO_KEY, _SHORT_PRESS)
        else:
            return_text = miva.press_key(_INFO_KEY, _SHORT_PRESS)
        if len(return_text) == 4 and return_text[0:2] == 'KE':
            print('key: Info ({})'.format(re_match_result[7]))
    elif re.match(r'(:)(key)(:)(run|stop)', cmd):
        # Send [RUN | STOP] key command [Short | Long] Press
        re_match_result = re.match(r'(:)(key)(:)(run|stop)((:)(short|long))?', cmd)
        if re_match_result[4] == 'run':
            return_text = miva.press_key(_RUN_KEY, _MID_PRESS)
        elif re_match_result[4] == 'stop':
            return_text = miva.press_key(_RUN_KEY, _MID_PRESS)
        if len(return_text) == 4 and return_text[0:2] == 'KE':
            print('key: {}'.format(re_match_result[4].upper()))
    elif re.match(r'(:)(key)(:)(ok|ent(er)?)', cmd):
        # Send [OK | ENTER] key command [Short] Press
        return_text = miva.press_key(_OK_KEY, _SHORT_PRESS)
        if len(return_text) == 4 and return_text[0:2] == 'KE':
            print('key: Ok/Enter (short)')
    elif re.match(r'(:)(key)(:)(up)', cmd):
        # Send [UP] key command [Short] Press
        return_text = miva.press_key(_UP_KEY, _SHORT_PRESS)
        if len(return_text) == 4 and return_text[0:2] == 'KE':
            print('key: Up (short)')
    elif re.match(r'(:)(key)(:)(down)', cmd):
        # Send [DOWN] key command [Short] Press
        return_text = miva.press_key(_DOWN_KEY, _SHORT_PRESS)
        if len(return_text) == 4 and return_text[0:2] == 'KE':
            print('key: Down (short)')
    elif re.match(r'(:)(key)(:)(bol(us)?)', cmd):
        # Send [BOLUS] key command [Short] Press
        return_text = miva.press_key(_BOLUS_KEY, _SHORT_PRESS)
        if len(return_text) == 4 and return_text[0:2] == 'KE':
            print('key: Bolus (short)')
    elif re.match(r'(:)(key)(:)(power)', cmd):
        # Send [ON/OFF] key command [Long] Press
        return_text = miva.press_key(_POWER_KEY, _LONG_PRESS)
        if len(return_text) == 4 and return_text[0:2] == 'KE':
            print('key: On/Off (long)')
    # ==============================================
    elif re.match(r'(:)(pin)(:)(\d{3})', cmd):
        # Input PIN number
        pin_number = re.match(r'(:)(pin)(:)(\d{3})', cmd)[4]
        print('Input PIN number [{}]...'.format(pin_number))
        time.sleep(_TIMEOUT * 10)
        #
        miva.input_digits(pin_number)
    elif re.match(r'(:)(cloc(k)?)(\?)', cmd):
        # Read Pump Time
        pump_time_str = miva.read_pump_time()
        if len(pump_time_str) == 10:
            # print(pump_time_str)
            day = str(int(pump_time_str[0:2], 16)).zfill(2)
            month = str(int(pump_time_str[2:4], 16)).zfill(2)
            year = str(int(pump_time_str[4:6], 16) + 2000).zfill(2)
            hour = str(int(pump_time_str[6:8], 16)).zfill(2)
            minute = str(int(pump_time_str[8:10], 16)).zfill(2)
            second = str(int(pump_time_str[10:12], 16)).zfill(2)
            print('{0}-{1}-{2} {3}:{4}:{5}'.format(year, month, day, hour, minute, second))
            print("Read pump time - Done!")
            return_msg = '{0}-{1}-{2} {3}:{4}:{5}'.format(year, month, day, hour, minute, second)
        elif len(pump_time_str) == 8:
            # The pump time is in relative format
            # [pump_time_str] is in format 'F0 00 00 00', So the first 'F' need to be removed.
            pump_life = int('0' + pump_time_str[1:], 16)
            active_since = datetime.datetime.now() - datetime.timedelta(seconds=pump_life)
            active_since_str = active_since.strftime("%Y-%m-%d %H:%M:%S")
            pump_life_delta = datetime.timedelta(seconds=pump_life)
            print('{0} (0x{1} - {2} sec - {3})'.format(active_since_str, \
                                                       pump_time_str, \
                                                       pump_life, \
                                                       pump_life_delta))
            print("Read pump time - Done!")
        else:
            print('Error: read pump time failed')
    elif re.match(r'(:)(cloc(k)?)(:)(sync(hronize)?)', cmd):
        # Synchronize Pump Time with PC time
        status = miva.write_pump_time()
        if status:
            current_date_time = datetime.datetime.now()
            current_date = datetime.date(current_date_time.year, \
                    current_date_time.month, current_date_time.day)
            current_time = datetime.time(current_date_time.hour, \
                    current_date_time.minute, current_date_time.second)
            print('{0} {1}'.format(current_date, current_time))
            print("Write pump time - Done!")
        else:
            print('Error: write pump time failed')
    elif re.match(r'(\*rst)', cmd):
        # Reset Pump Total VINF and Total ON Time
        print("Reset pump ...")
        status = miva.reset_pump()
        if status:
            print("Reset service total volume infused - Done!")
            print("Reset service total ON time - Done!")
        else:
            print('Error: reset pump failed')
    elif re.match(r'(:)(raw)(:)(.+)', cmd):
        # Send Raw Command (the command that pump recognize directly)
        re_match_result = re.match(r'(:)(raw)(:)(.+)', cmd)
        sent_text = re_match_result[4].upper().strip(' \t\r\n\0')
        print('Command to send: {}'.format(sent_text))
        return_msg = miva.query(sent_text)
        print(return_msg)
    elif re.match(r'(:)(plat(form)?)(\?)', cmd):
        # Read Platform (H and K pump only)
        platform = miva.read_platform()
        if platform != '':
            print('pump platform: [{}]'.format(platform))
            return_msg = platform
        else:
            print('Error: read pump platform failed')
    elif re.match(r'(:)(plat(form)?)(\s)(h|k)', cmd):
        # Write Platform (H and K pump only)
        re_match_result = re.match(r'(:)(plat(form)?)(\s)(h|k)', cmd)
        platform = re_match_result[5]
        status = miva.write_platform(platform)
        if status:
            print("Write pump platform - Done!")
        else:
            print('Error: write pump platform failed')
    # ==============================================
    elif re.match(r'(:)(sim(ulator)?)(:)(lib(rary)?)(:)(load)(\s)(.+)', cmd):
        # Load Library (to the pump simulator)
        re_match_result = re.match(r'(:)(sim(ulator)?)(:)(lib(rary)?)(:)(load)(\s)(.+)', cmd)
        lib_path = re_match_result[10]
        if path.exists(lib_path):
            library.load(lib_path)
            miva.pump_config["library_path"] = lib_path
            json_to_file(miva.pump_config, miva.pump_config_path)
            print("library: ")
            # Title
            print("    {0:5s}    |    {1:5s}    |    {2:15s}    "
                    .format('id', 'ver.', 'name'))
            print("----{0:5s}----|----{1:5s}----|----{2:15s}----"\
                    .format('-----', '-----', '---------------'))
            # Content
            print("    {0:5s}    |    {1:5s}    |    {2:15s}"\
                    .format(str(library.get_id()), \
                            str(library.get_version()), \
                            library.get_name()))
    elif re.match(r'(:)(sim(ulator)?)(:)(lib(rary)?)(:)(load)(\?)', cmd):
        # Query Library Name (from the pump simulator)
        print("library: ")
        # Title
        print("    {0:5s}    |    {1:5s}    |    {2:15s}    "
                .format('id', 'ver.', 'name'))
        print("----{0:5s}----|----{1:5s}----|----{2:15s}----"\
                .format('-----', '-----', '---------------'))
        # Content
        print("    {0:5s}    |    {1:5s}    |    {2:15s}"\
                .format(str(library.get_id()), \
                        str(library.get_version()), \
                        library.get_name()))
        lib_info = {'id':library.get_id(), \
                    'version':library.get_version(), \
                    'name':library.get_name()
                    }
        return_msg = json.dumps(lib_info)
    elif re.match(r'(:)(sim(ulator)?)(:)(prot(ocol)?)(:)(list)(\?)', cmd):
        # Query Protocol List (from the pump simulator)
        protocol_list = []
        protocols = library.get_protocols()
        for each_protocol in protocols:
            protocol_info = {'id':str(each_protocol['id']),
                             'name': str(each_protocol['content']['name']),
                             'deliveryMode': str(each_protocol['content']['deliveryMode'])}
            protocol_list.append(protocol_info)
        return_msg = json.dumps(protocol_list)
        print(json.dumps(protocol_list, indent=4))
        print('Total [{}] protocols extracted'.format(len(protocols)))
    elif re.match(r'(:)(sim(ulator)?)(:)(prot(ocol)?)(:)(sel(ect)?)(\s)(.+)', cmd):
        # Select Protocol (on the pump simulator and on the pump)
        re_match_result = re.match(r'(:)(sim(ulator)?)(:)(prot(ocol)?)(:)(sel(ect)?)(\s)(.+)', \
                                   cmd_original, re.IGNORECASE)
        protocol_name = re_match_result[11]
        # print('protocol name = {}'.format(protocol_name))
        # MUST BE DEEP COPY
        protocol = copy.deepcopy(library.get_protocol(protocol_name))
        if protocol is not None:
            # Quit the current infusion task on the simulator
            if 'PAUSED' in infusion.status.name:
                infusion_monitor.set_command('quit')
            elif 'RUNNING' in infusion.status.name:
                print('Abort: infusion is running...')
                return return_msg
            # select protocol on the pump
            miva.select_protocol(library, protocol)
            # set protocol on the simulator
            infusion.set_protocol(protocol)
            # save selected protocol
            miva.pump_config["protocol_selected"] = protocol_name
            json_to_file(miva.pump_config, miva.pump_config_path)
    elif re.match(r'(:)(sim(ulator)?)(:)(prot(ocol)?)(:)(sel(ect)?)(\?)', cmd):
        # Query the Selected Protocol (on the pump simulator and on the pump)
        protocol_name = protocol['content']['name']
        protocol_id = protocol['id']
        protocol_info = {'id':protocol_id, 'name':protocol_name} 
        return_msg = json.dumps(protocol_info)
    elif re.match(r'(:)(sim(ulator)?)(:)(prot(ocol)?)(:)(para(meter)?)(:)(list)(\?)', cmd):
        # Print Parameter List
        parameter_list = get_parameter_list(library, protocol)
        return_msg = ' '.join(parameter_list)
        print('==')
        print('Infusion Parameter List = {}'.format(parameter_list))
    elif re.match(r'(:)(sim(ulator)?)(:)(prot(ocol)?)(:)(para(meter)?)(:)(.+)(\s)(\d+)((.)(\d+))?', cmd):
        # Change Parameter Value (on the pump simulator and on the pump)
        re_match_result = re.match(r'(:)(sim(ulator)?)(:)(prot(ocol)?)(:)(para(meter)?)(:)(.+)(\s)(\d+)((.)(\d+))?', \
                                   cmd_original, re.IGNORECASE)
        parameter_name = re_match_result[11]
        parameter_value = re_match_result[13]
        parameter_list = get_parameter_list(library, protocol)
        if parameter_name not in parameter_list \
                and parameter_name not in ['drugAmount', 'diluteVolume']:
            print('    Invalid parameter name: {}'.format(parameter_name))
            return return_msg
        if parameter_name in ['drug', 'label']:
            print('    Abort: [{}] is NOT editable'.format(parameter_name))
            return return_msg
        if parameter_name == 'concentration':
            if protocol['content']['drug']['content']['concentrationImmutable']:
                print('    Abort: concentration is NOT editable')
            else:
                print('    Abort: concentration is NOT editable'.format(parameter_name))
                print('           type \'drugAmount\' or \'diluteVolume\' instead')
            return return_msg
        # get parameter original, maximum, minimum values and unit
        parameter = get_parameter(protocol, parameter_name, infusion)
        original_value = float(parameter['value'])
        max_value = float(parameter['max'])
        min_value = float(parameter['min'])
        upper_limit = max_value
        lower_limit = min_value
        unit = parameter['unit']
        # get NEW value
        if parameter['name'] == 'weight':
            upper_limit = float(parameter['upper_limit'])
            lower_limit = float(parameter['lower_limit'])
            print('{0} = {1:.2f} {2} (max: {3:.2f} {2}; min: {4:.2f} {2})'\
                    .format(parameter_name, original_value, \
                            unit, upper_limit, lower_limit))
        elif parameter['name'] == 'diluteVolume':
            print('{0} = {1:.0f} {2} (max: {3:.0f} {2}; min: {4:.0f} {2})'\
                    .format(parameter_name, original_value, \
                            unit, max_value, min_value))
        else:
            print('{0} = {1:.2f} {2} (max: {3:.2f} {2}; min: {4:.2f} {2})'\
                    .format(parameter_name, original_value, \
                            unit, max_value, min_value))
        if parameter['name'] == 'diluteVolume':
            new_value_str = parameter_value
            new_value_str = int(float(new_value_str))
            new_value_str = str(new_value_str)
        else: 
            new_value_str = parameter_value
        # check if it is a number
        if new_value_str.replace('.', '', 1).isdigit():
            if float(new_value_str) > max_value:
                print('Out of range: {0} will be set to {1:.2f} {2}'.format(parameter_name, max_value, unit))
                new_value = max_value
            elif float(new_value_str) < min_value:
                print('Out of range: {0} will be set to {1:.2f} {2}'.format(parameter_name, min_value, unit))
                new_value = min_value
            else:
                new_value = float(new_value_str)
        elif new_value_str.lower().strip(' \t\r\n\0') == 'max':
            if parameter['name'] == 'weight':
                new_value = upper_limit
            else:
                new_value = max_value
        elif new_value_str.lower().strip(' \t\r\n\0') == 'min':
            if parameter['name'] == 'weight':
                new_value = lower_limit
            else:
                new_value = min_value
        else:
            print('invalid input: {}'.format(new_value_str))
            return return_msg
        if new_value == original_value:
            print('Abort: two values are equal')
            return return_msg
        # select [Parameter] on the pump
        result = miva.select_parameter(library, protocol, parameter_name)
        if result:
            print('==')
            print('Set [{0}] to [{1:.2f}] {2}...'.format(parameter_name, new_value, unit))
            time.sleep(_TIMEOUT * 10)
            # Set [Parameter] Value on the Pump
            miva.set_parameter_value(new_value, parameter)
            # Update [Parameter] value
            if parameter_name == 'weight':
                infusion.patient_weight = new_value
            elif parameter_name in ['drugAmount', 'diluteVolume']:
                protocol['content']['drug']['content'][parameter_name]['value'] = new_value
            else:
                protocol['content']['program'][parameter_name]['value'] = new_value
            #
            infusion.set_protocol(protocol)
        # Return to top of the [Parameter] list
        miva.return_top(library, protocol, parameter_name)
        # Display infusion parameters
        infusion.print_parameters()
    elif re.match(r'(:)(sim(ulator)?)(:)(prot(ocol)?)(:)(para(meter)?)(:)((?!list).+)(\?)', cmd):
        # Query Parameter Value (from the pump simulator)
        re_match_result = re.match(r'(:)(sim(ulator)?)(:)(prot(ocol)?)(:)(para(meter)?)(:)((?!list).+)(\?)', \
                                   cmd_original, re.IGNORECASE)
        parameter_name = re_match_result[11]
        parameter_list = get_parameter_list(library, protocol)
        if parameter_name not in parameter_list \
                and parameter_name not in ['drugAmount', 'diluteVolume']:
            print('    Invalid parameter name: {}'.format(parameter_name))
            return return_msg
        if parameter_name == 'drug':
            drug_info = protocol['content']['drug']['content']
            return_msg = json.dumps(drug_info)
            return return_msg
        if parameter_name == 'label':
            label_set = library.get_protocol_label_set(protocol['content']['name'])
            return_msg = json.dumps(label_set)
            return return_msg
        if parameter_name == 'concentration':
            if protocol['content']['drug']['content']['concentrationImmutable']:
                print('    Abort: concentration is NOT editable')
            else:
                print('    Abort: concentration is NOT editable'.format(parameter_name))
                print('           type \'drugAmount\' or \'diluteVolume\' instead')
            return return_msg
        # get parameter original, maximum, minimum values and unit
        parameter = get_parameter(protocol, parameter_name, infusion)
        return_msg = json.dumps(parameter)
        print(json.dumps(parameter, indent=4))
    elif re.match(r'(:)(sim(ulator)?)(:)(inf(usion)?)(:)(run)', cmd):
        # Run Infusion (on the pump)
        if library.get_library() is not None:
            if protocol is not None:
                # protocl_name = protocol['content']['name']
                infusion.set_protocol(protocol)
                print("Protocol: ")
                print()
                # Title
                print("    {0:5s}    |    {1:15s}    |    {2:20s}"\
                        .format('id', 'name', 'mode'))
                print("----{0:5s}----|----{1:15s}----|----{2:20s}"\
                        .format('-----', '---------------', '--------------------'))
                # Content
                print("    {:5s}    |".format(str(protocol['id'])), end='')
                print("    {:15s}    |".format(protocol['content']['name']), end='')
                print("    {:20s}".format(protocol['content']['deliveryMode']))
                print()
                #
                print("Extracted Parameters:")
                print()
                if protocol['content']['deliveryMode'] == 'continuousInfusion':
                    infusion.print_cont_parameters()
                if protocol['content']['deliveryMode'] == 'bolusInfusion':
                    infusion.print_bolus_parameters()
                if protocol['content']['deliveryMode'] == 'intermittentInfusion':
                    infusion.print_int_parameters()
                #
                start_yes_no = 'y'
                if start_yes_no.lower().find('y') != -1 or start_yes_no == '':
                    print('infusion start...')
                    # Run Protocol on the Pump
                    miva.run_protocol(library, protocol)
                    # infusion.run_infusion()
                    if infusion.status in [InfusionStatus.PAUSED_AUTO_BOLUS, \
                                           InfusionStatus.PAUSED_DEMAND_BOLUS, \
                                           InfusionStatus.PAUSED_LOADING_DOSE, \
                                           InfusionStatus.PAUSED_CLINICIAN_DOSE]:
                        infusion_monitor.output_enabled = False
                        #
                        resume_dose = 'y'
                        if resume_dose == 'y':
                            print('==')
                            print('resume dose...')
                            time.sleep(_TIMEOUT * 10)
                            # Press [Ok] Key
                            return_text = miva.press_key(_OK_KEY, _SHORT_PRESS)
                            if len(return_text) == 4 and return_text[0:2] == 'KE':
                                print('key: Ok/Enter (short)')
                            time.sleep(_TIMEOUT * 10)
                        else:
                            infusion.status = InfusionStatus.PAUSED
                            basal_rate = infusion.get_rate(infusion.protocol)
                            infusion.motor.set_rate(basal_rate)
                            print('==')
                            print('cancel dose...')
                            time.sleep(_TIMEOUT * 10)
                            # Press [Down] Key
                            return_text = miva.press_key(_DOWN_KEY, _SHORT_PRESS)
                            if len(return_text) == 4 and return_text[0:2] == 'KE':
                                print('key: Down (short)')
                            time.sleep(_TIMEOUT * 10) 
                            # Press [Ok] Key
                            return_text = miva.press_key(_OK_KEY, _SHORT_PRESS)
                            if len(return_text) == 4 and return_text[0:2] == 'KE':
                                print('key: Ok/Enter (short)')
                            time.sleep(_TIMEOUT * 10)
                        infusion_monitor.output_enabled = True
                            
                else:
                    print('infusion aborted...')
            else:
                print('Empty Protocol: Use <SP> command to select protocol first.')
        else:
            print('Empty Library: Use <LL> command to load library first.')
    elif re.match(r'(:)(sim(ulator)?)(:)(inf(usion)?)(:)(stop)', cmd):
        # Stop Infusion (on the pump)
        return_text = miva.trigger_stop_infusion()
        if return_text:
            print('Trigger Stop Infusion')
    elif re.match(r'(:)(sim(ulator)?)(:)(inf(usion)?)(:)(cli(nician)?)(:)(run)', cmd):
        # Run [Clinician Dose] on the Pump
        clinician_code = library.get_access_code('clinician')
        switches = protocol['content']['program']['switches']
        if switches.get('clinicianDose') and switches['clinicianDose']:
            # Get the platform of the pump
            platform = miva.read_platform()
            if platform in ['H', 'K']:
                if infusion.status in [InfusionStatus.RUNNING_AUTO_BOLUS, \
                                       InfusionStatus.RUNNING_DEMAND_BOLUS, \
                                       InfusionStatus.RUNNING_LOADING_DOSE]:
                    print('clinician dose NOT granted: infusion status = {}'\
                          .format(infusion.status.name))
                    print()
                else:
                    miva.run_clinician_dose(clinician_code)
            elif platform == 'C':
                if infusion.status in [InfusionStatus.PAUSED_MPH_REACHED, \
                                       InfusionStatus.PAUSED_MPI_REACHED] \
                        and infusion.previous_status in [InfusionStatus.RUNNING_AUTO_BOLUS, \
                                                         InfusionStatus.RUNNING_DEMAND_BOLUS, \
                                                         InfusionStatus.RUNNING_LOADING_DOSE]:
                    print('clinician dose NOT granted: previous infusion status = {}'\
                          .format(infusion.previous_status.name))
                    print()
                elif infusion.status in [InfusionStatus.RUNNING_AUTO_BOLUS, \
                                       InfusionStatus.RUNNING_DEMAND_BOLUS, \
                                       InfusionStatus.RUNNING_LOADING_DOSE]:
                    print('clinician dose NOT granted: infusion status = {}'\
                          .format(infusion.status.name))
                    print()
                else:
                    miva.run_clinician_dose(clinician_code)
    elif re.match(r'(:)(sim(ulator)?)(:)(inf(usion)?)(:)(stat(us)?)(\?)', cmd):
        # Check Infusion Status
        infusion_monitor.set_command('status')
        return_msg = infusion.status.name
    elif re.match(r'(:)(sim(ulator)?)(:)(inf(usion)?)(:)(para(meter)?)(\?)', cmd):
        # Query Infusion Runtime Parameters
        return_msg = json.dumps(socket_server.infusion.get_runtime_parameters())
        print(json.dumps(socket_server.infusion.get_runtime_parameters(), indent=4))
    if re.match(r'(:)(sim(ulator)?)(:)(inf(usion)?)(:)(quit|pause|parameters)', cmd):
        # Run [Infusion Monitor] Command
        re_match_result = re.match(r'(:)(sim(ulator)?)(:)(inf(usion)?)(:)(quit|pause|parameters)', cmd)
        infusion_monitor.set_command(re_match_result[8])
    # ==============================================
    # Resume [Event Log Monitor]
    if event_log_monitor.is_on():
        event_log_monitor.resume()
    # ==============================================
    # Enable [Event Log Monitor] Output
    infusion_monitor.enable_output()
                    
    return return_msg


def main():
    '''main function'''
    try:
        hostname = socket.gethostname()
        local_ip = socket.gethostbyname(hostname)
        print("IP address: {0}".format(local_ip))
    
        # HOST = '127.0.0.1'  # Standard loopback interface address (localhost)
        HOST = local_ip
        PORT = 65432  # Port to listen on (non-privileged ports are > 1023)
        # Create the server, binding to localhost on port 9999
#         with socketserver.TCPServer((HOST, PORT), SocketRequestHandler) as tcp_server:
#             #######################################
#             # create a [request_handler] thread   #
#             #######################################
#             request_handler_thread = Thread(target=tcp_server.serve_forever, args=())
#             request_handler_thread.daemon = True
#             request_handler_thread.start()
#             request_handler_thread.join()
        with SocketServer((HOST, PORT), SocketRequestHandler,) as socket_server:
            socket_server.start()
    except KeyboardInterrupt:
        print('Exiting: [Ctrl + C] is pressed')
        pass


if __name__ == "__main__":
    main()
